#lang typed/racket

(provide string->program)

(require "types.rkt")
(require threading)

(: string->program (-> String Program))
(define (string->program src)
  (~> src
      string->list
      tokenize
      parse))

(: tokenize (-> (Listof Char) (Listof Token)))
(define (tokenize chars)

  (: parse-char (-> Char (U Token Null)))
  (define (parse-char c)
    (case c
      [(#\>) 'pointer-increment]
      [(#\<) 'pointer-decrement]
      [(#\+) 'increment]
      [(#\-) 'decrement]
      [(#\.) 'output]
      [(#\,) 'input]
      [(#\[) 'loop-start]
      [(#\]) 'loop-end]
      [else '()]))
  (cast (filter (lambda (x) (not (null? x)))
                (map parse-char chars)) (Listof Token)))

(: parse (-> (Listof Token) Program))
(define (parse tokens)
  (: safe-parse (-> (Listof Token) Program (Listof Loop) Program))
  (define (safe-parse tokens acc loop-stack)
    (if (null? tokens)
        (reverse acc)
        (inner-parse tokens acc loop-stack)))

  (: inner-parse (-> (Listof Token) Program (Listof Loop) Program))
  (define (inner-parse tokens acc loop-stack)
    (define next (car tokens))

    (: continue (-> Program (Listof Loop) Program))
    (define (continue acc loop-stack)
      (safe-parse (cdr tokens) acc loop-stack))

    (define (new-loop)
      (continue acc (cons '() loop-stack)))

    (define (end-loop)
      (continue (cons (reverse (car loop-stack)) acc)
                (cdr loop-stack)))

    (define (add-next)
      (if (null? loop-stack)
          (continue (cons (cast next Instruction) acc) loop-stack)
          (continue acc (cons (cons (cast next Instruction) (car loop-stack))
                              (cdr loop-stack)))))

    (case next
      ['loop-start (new-loop)]
      ['loop-end (end-loop)]
      [else (add-next)]))

  (safe-parse tokens '() '()))
